#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
"""
    dnsqo is a tool for mass reverse DNS resolution and name resolution brute
    forcing (resolving all words dot domain name) with many customizable
    parameters which can help to evade some intrusion detection
    and prevention systems (IDS/IPS). dnsqo can be used for testing the
    security of a DNS server's data, and see how well the IDS/IPS is working
    against it.

    features:

	[o] using a lot of random factors to harden the detection of tool
	    activity by IDS/IPS:

	    - waits for a random time within a range specified by user between
	      sending queries

	    - packet ID and source port forceful randomization, - some OS's
	      don't do this by default

	    - brute forcer and mass reverse name resolution queries are
	      not sequential - they sent in random order too.

	    - each nameserver for each query is also randomly chosen from
	      list/file. You can make your own files with nameservers and
	      try to use PublicNameServers (domain name servers that provide
	      DNS resolutions for anyone who requests it from servers)

	[o] multiple PTR records support: while most Reverse DNS lookup
	    entries only have one PTR record, it is perfectly legal to have
	    many different PTR records. for example, if a webserver
	    supports many virtual hosts, there can be one PTR record for each
	    host and some versions of name server software will automatically
	    add a PTR record for each host. multiple PTR records can cause
	    a couple of problems, including triggering bugs in programs that
	    only expect there to ever be a single PTR record and, in the case
	    of a large webserver, having hundreds of PTR records can cause
	    the DNS packets to be much larger than normal.

	[o] no multithreading, because huge amount of dns-requests are
	    easily recognized as an 'attack'  

	[o] don't do zonetransfer, because we can do it in a more professional
	    manner with "host" tool. 

    requirements:

	python: http://www.python.org, dnsqo requires python-2.3 or later,
	and is primarily developed against python-2.5
	
	dnspython: http://www.dnspython.org, a DNS toolkit for Python

    addons:

	ns.lst - default dnsqo nameservers file
	iplistgenq.py - IP list generator
	w47988.lst - wordlist with all 47988 alphanumeric 3 chars length words
	w4400.lst - wordlist with 4402 common words longer than 3 chars

    examples:

	dnsqo -r 192.168.1.1,10.0.0.33:5300 192.168.0.0-254 -o=results.log
	    do reverse DNS resolution on all IP adresses from 192.168.0.0
	    to 192.168.0.254 using 192.168.1.1 and 10.0.0.33(with domain
	    name service on 5300 port) as nameservers and save output
	    to results.log

	dnsqo -r -nsl=ns2.lst -ipl=net.lst --src_port_rnd_off --randomize_ip_off
	    do reverse DNS resolution on IP adresses from net.lst using
	    nameservers from ns2.lst, don't use random source port in queries,
	    turn OFF random IP selection from an IP address list when doing
	    queries and save output to default 'dnsqo.log' file.

	dnsqo -l -nsl=ns.lst mydomain.my an1-3.lst --timeout=5
	    do brute force name resolution from 'an1-3.lst' wordlist
	    in 'mydomain.my' domain using nameservers from default dnsqo
	    nameservers file - 'ns.lst', set time to wait before the query
	    will time out in 5 seconds.
	
	
	^_^
"""

import sys
from string import join
import random
import dns.resolver
import dns.reversename
import time
import datetime
from socket import inet_aton

__version__ = 'dnsqo.000'

def usage():
    print """
                                                                [%s]

 do reverse-DNS resolution on the given hosts to get their names:
 -r <nameserver> <hosts>

	<nameserver>: nameserver1_ip[:port][,nameserver2_ip],...
	or use '-nsl=<filename>' to read the nameservers from <filename>
	* see default dnsqo nameservers file 'ns.lst' for more details

	<hosts> can be the IP address range like 192.168.0.1-254 or
	you can specify a file with IP adresses that way:
	-ipl=<filename>. file format is - one IP address per line,
	* you can use iplistgenq.py to generate some

  --randomize_ip_off: turn OFF random IP selection from a list or range,
    when doing queries, default is ON


 do brute force on name resolution:
 -l <nameserver> <domain> <wordlist>

	<nameserver>: nameserver1_ip[:port][,nameserver2_ip],...
	or use '-nsl=<filename>' to read the nameservers from <filename>
	* see default dnsqo nameservers file 'ns.lst' for more details

  --randomize_names_off: turn off random selection of name-to-resolve
    from wordlist, default is ON


 common options:

  --timeout=<x>: time to wait before the query times out, default is 3 s.
  --retries=<x>: the number of retries for a lookup, default is 5.
  --delay_interval=<x-y>: wait for a random time within x..y range
    between sending queries, default is 0-3 s.
  --use_tcp: use TCP to transport DNS messages, default is UDP
  --src_port_rnd_off: turn the source port randomization OFF, default is ON
  -o=<filename>: append output to a file, default is 'dnsqo.log'
  -v: turn on verbose mode
  -h: print this help page

 see the source for details""" % __version__
    sys.exit(0)



#nondevided = ip_addr[:port][,ip_addr[:port]] ...
def ipsnports_to_tuple(nondevided):
    nameservers = []
    for pair in nondevided.split(','):
	a = pair.split(':')
	try:
	    if a[1]: nameservers.append((a[0], int(a[1])))
	except:
	    nameservers.append((pair,53)) # 53 is the default dns port
    return nameservers



# Yet Another Unique() Function by Jordan Callicoat
def unique(inlist, keepstr=True):
    typ = type(inlist)
    if not typ == list:
	inlist = list(inlist)
    i = 0
    while i < len(inlist):
	try:
	    del inlist[inlist.index(inlist[i], i + 1)]
	except:
	    i += 1
    if not typ in (str, unicode):
	inlist = typ(inlist)
    else:
	if keepstr:
	    inlist = ''.join(inlist)
    return inlist



def randomize_a_list(lst):
    for i in range(len(lst)-1,0,-1):
	choice = int(random.randint(0,i))
	lst[choice], lst[i] = lst[i], lst[choice]
    return lst



def dnsqo_lookups_by_wordlist(nameservers, domain, wwlist, timeout, \
	retries, delay_interval, query_type, randomize_src_port, verbose):

    domain = '.'+domain
    results = [] # name, ip
    src_port = 0
    print ''
    tretries = retries

    for word in wwlist:

	time.sleep(random.uniform(delay_interval[0],delay_interval[1]))

	hostname = word+domain
	reque = dns.message.make_query(hostname,'A',1)
	retries = tretries
	while retries != 0:
	    stime = datetime.datetime.now().time()
#	repair .000 microseconds bug
	    if stime.microsecond == 0: stime = str(stime)+'.000'
	    else: stime = str(stime)

#	the Registered Ports are those in the range 102449151,
#	the Dynamic and/or Private Ports are those in the range 4915265535
	    if randomize_src_port: src_port = random.randrange(1024,65535)

	    ns = nameservers[random.randrange(0,len(nameservers))]
	
	    try:
		if query_type == 0:
		    n = dns.query.udp(reque,ns[0],timeout,ns[1],2,'0.0.0.0',src_port) # 2 = AF_INET
		else:
		    n = dns.query.tcp(reque,ns[0],timeout,ns[1],2,'0.0.0.0',src_port)
	    except (dns.exception.Timeout):
		print " %s --> %s : %s --- [x] timeout" % (stime[:12],ns[0],hostname)
		retries -= 1
	    except (dns.query.BadResponse):
		print " %s --> %s : %s --- [x] bad response" % (stime[:12],ns[0],hostname)
		retries -= 1
	    except (dns.query.UnexpectedSource):
		print " %s --> %s : %s --- [x] unexpected source" % (stime[:12],ns[0],hostname)
		retries -= 1
	    except: 
		if sys.exc_value[0] == 10048: # if our source port is already used
		    pass
		else:
		    print sys.exc_type, sys.exc_value
		    retries -= 1
	    else:
		if n.answer:
		    for mm in n.answer:
			for aak in mm.to_text().split('\n'):
			    aak = aak.split()
			    if verbose: print " %s --> %s :" % (stime[:12],ns[0]),
			    if str(aak[3]) == 'CNAME':
#				CNAME record, a record in a DNS database that 
#				indicates the true, or canonical, host name of a computer
#				that its aliases are associated with
				if verbose: print "%s ---CNAME--> %s" % (hostname, aak[4][:-1])
				results.append((hostname, aak[4][:-1]))
			    else:
				if verbose: print "%s %s" % (aak[0][:-1], aak[4])
				results.append((aak[0][:-1], aak[4]))
		else:
		    if verbose: print " %s --> %s : %s --- no such name" % (stime[:12],ns[0],hostname)
		retries = 0
    return results



def dnsqo_reverse_lookups(nameservers, ip_list, timeout, retries, delay_interval,\
	query_type, randomize_src_port, output_file, verbose):

    results = []
    print ''
    tretries = retries
    src_port = 0

    for adre in ip_list:
	time.sleep(random.uniform(delay_interval[0],delay_interval[1]))
	reque = dns.message.make_query(dns.reversename.from_address(adre),'PTR',1)
	retries = tretries
	while retries != 0:
	    stime = datetime.datetime.now().time()
#	    repair .000 microseconds bug
	    if stime.microsecond == 0: stime = str(stime)+'.000'
	    else: stime = str(stime)

#	    the Registered Ports are those in the range 102449151,
#	    the Dynamic and/or Private Ports are those in the range 4915265535
	    if randomize_src_port: src_port = random.randrange(1024,65535)

	    ns = nameservers[random.randrange(0,len(nameservers))]
	    try:
		if query_type == 0:
		    n = dns.query.udp(reque,ns[0],timeout,ns[1],2,'0.0.0.0',src_port) # 2 = AF_INET
		else:
		    n = dns.query.tcp(reque,ns[0],timeout,ns[1],2,'0.0.0.0',src_port)
	    except (dns.exception.Timeout):
		print " %s --> %s : %s --- [x] timeout" % (stime[:12],ns[0],adre)
		retries -= 1
	    except (dns.query.BadResponse):
		print " %s --> %s : %s --- [x] bad response" % (stime[:12],ns[0],adre)
		retries -= 1
	    except (dns.query.UnexpectedSource):
		print " %s --> %s : %s --- [x] unexpected source" % (stime[:12],ns[0],adre)
		retries -= 1
	    except: 
		if sys.exc_value[0] == 10048: # if our source port is already used
		    pass
		else:
		    print sys.exc_type, sys.exc_value
		    retries -= 1
	    else:
		if n.answer:
		    if verbose: print " %s --> %s : %s" % (stime[:12],ns[0],adre),
		    for a in n.answer[0].to_text().split('\n'): # multiple PTR records
			if verbose: print a.split(" ")[4][:-1],
			results.append((adre,a.split(" ")[4][:-1]))
		    if verbose: print ''
		else:
		    if verbose: print " %s --> %s : %s --- no such name" % (stime[:12],ns[0],adre)
		retries = 0
    return results



def main():

#    no more validation of a data submitted by user, who is responsible
#    for correctness of submitted arguments and options
    if len(sys.argv) < 2: usage() 

    verbose = False
    timeout = 3 # the query timeout in seconds
    retries = 5 # the number of retries for a lookup
    delay_interval = [0,3]
    query_type = 0 # 0 = udp, 1 = tcp
    randomize_src_port = True
    output_file = 'dnsqo.log' # default output file name

    if '-h' in sys.argv: usage()
    if '-v' in sys.argv: verbose = True
    if '--use_tcp' in sys.argv: query_type = 1
    if '--src_port_rnd_off' in sys.argv: randomize_src_port = False
    for o in sys.argv:
	if '-o=' in o: output_file=o[3:]
	elif '--delay_interval=' in o: delay_interval = (int(o[17:].split('-')[0]),int(o[17:].split('-')[1]))
	elif '--timeout=' in o: timeout = int(o[10:])
	elif '--retries=' in o: retries = int(o[10:])

    if sys.argv[1] == '-r':
	if '-nsl=' in sys.argv[2]:
	    nsl = open(sys.argv[2][5:], 'r')
	    nsff = []
	    for line in nsl:
		if line[0] != '#': nsff.append(line.split('#')[0].strip(' \n\t'))
	    nsl.close()
	    ns_tuple = ipsnports_to_tuple(",".join(nsff))
	else:
	    ns_tuple = ipsnports_to_tuple(sys.argv[2])

        ip_list = []
	if '-ipl=' in sys.argv[3]:
	    ipl = open(sys.argv[3][5:], 'r')
	    for line in ipl: ip_list.append(line.strip(' \n\t'))
	    ipl.close()
	else:
	    tmpip = sys.argv[3].split('.')
	    range_ip_prefics = "%s.%s.%s." %(tmpip[0],tmpip[1],tmpip[2])
	    for i in range(int(tmpip[-1].split('-')[-2]), int(sys.argv[3].split('-')[-1])+1):
		ip_list.append(range_ip_prefics+str(i))

	if not '--randomize_ip_off' in sys.argv:
	    ip_list = randomize_a_list(ip_list)

	ip_list = dnsqo_reverse_lookups(ns_tuple, ip_list, timeout, retries, delay_interval, \
		query_type, randomize_src_port, output_file, verbose)
	
	ip_list.sort(lambda x,y: cmp(inet_aton(x[0]),inet_aton(y[0])))
	
	log_file = open(output_file,'a')
	for ii in ip_list: log_file.write("%s %s\n" % (ii[0],ii[1]))
	log_file.close()

    elif sys.argv[1] == '-l':
	if '-nsl=' in sys.argv[2]:
	    nsl = open(sys.argv[2][5:], 'r')
	    nsff = []
	    for line in nsl:
		if line[0] != '#': nsff.append(line.split('#')[0].strip(' \n\t'))
	    nsl.close()
	    ns_tuple = ipsnports_to_tuple(",".join(nsff))
	else:
	    ns_tuple = ipsnports_to_tuple(sys.argv[2])

	wwlist = []
	wlist = open(sys.argv[4], 'r')
	for wword in wlist: wwlist.append(wword.strip(' \n\t'))
	wlist.close()

	if not '--randomize_names_off' in sys.argv:
	    wwlist = randomize_a_list(wwlist)

	results = dnsqo_lookups_by_wordlist(ns_tuple, sys.argv[3], wwlist, timeout, \
		retries, delay_interval, query_type, randomize_src_port, verbose)

	results = unique(results)
	results.sort()

	log_file = open(output_file,'a')
	for ii in results: log_file.write("%s %s\n" % (ii[0],ii[1]))
	log_file.close()

    else: usage()

    return 0

if __name__== '__main__': main()
#404 -_^